import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

import 'mutation.dart';
import 'resolver.dart';
import 'state_container.dart';

abstract class Async<T> {}

class Uninitialized<T> extends Async<T> {}

class Loading<T> extends Async<T> {}

class Loaded<T> extends Async<T> {
  final T state;
  Loaded(this.state);
}

class Errored<T> extends Async<T> {
  final Exception exception;

  Errored(this.exception);
}

class AsyncResolverController<T> extends ResolverController<Async<T>> {
  AsyncResolverController(ValueNotifier<Async<T>> stateNotifier)
      : super(stateNotifier);

  void load(Future<T> Function() executor) {
    run(() async* {
      yield Loading<T>();
      final d = await executor();
      yield Loaded(d);
    });
  }
}

typedef AsyncResolverCallback<R, S, U> = Async<R> Function(
  AsyncResolverController<S> ctrl,
  Async<S> value,
  U? arg,
);

class AsyncMutationResolver<T extends Mutation, K>
    extends MutationResolver<T, Async<K>> {
  AsyncMutationResolver(
      AsyncResolverController<K> ctrl, AsyncResolverCallback<K, K, T> callback)
      : super(ctrl, (a, b, c) {
          return callback(a as AsyncResolverController<K>, b, c);
        });
}

class NotImplemented implements Exception {}

extension AsyncCast<T> on Future<T> {
  Future<Async<T>> toAsync() {
    return this
        .then((value) => Loaded<T>(value))
        .catchError((err) => throw Errored<T>(err));
  }
}

extension StateCast<T> on T {
  Loaded<T> ok() {
    return Loaded(this);
  }
}

extension AsyncError on Exception {
  Errored<T> err<T>() {
    return Errored<T>(this);
  }
}

class AsyncStateContainer<T> extends StateContainer<Async<T>> {
  Async<T> get initialState => Uninitialized<T>();

  Stream<Async<T>> init(context) async* {
    Future<Async<T>> req;

    try {
      req = load();
    } on NotImplemented {
      return;
    }

    yield Loading<T>();

    try {
      final d = await req;
      yield d;
    } on Exception catch (err) {
      yield Errored<T>(err);
    } on Errored<T> catch (err) {
      yield err;
    }
  }

  Future<Async<T>> load() {
    throw new NotImplemented();
  }

  @override
  AsyncResolverController<T> get resolverController =>
      AsyncResolverController(stateNotifier);

  @override
  void mutation<K extends Mutation>(AsyncResolverCallback<T, T, K> callback) {
    final mutationResolver =
        AsyncMutationResolver<K, T>(resolverController, callback);
    mutationResolvers[K] = mutationResolver;
  }
}
